/*  job_printintegralinfo.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_printintegralinfo.h"
#include <sstream>
#include "functions.h"
#include "int.h"
#include "sector.h"
#include "filedata.h"
#include "files.h"
#include "../yaml/include/yaml.h"
#include "topology.h"
#include "graph.h"

using namespace std;

namespace Reduze {

namespace {
JobProxy<PrintIntegralInfo> dummy;
}

bool PrintIntegralInfo::find_dependencies(const set<string>& outothers,//
		list<string>& in, list<string>& out, list<Job*>& auxjobs) {
	return true;
}

void PrintIntegralInfo::init() {
}

std::string PrintIntegralInfo::get_description() const {
	return "print integral info";
}

void PrintIntegralInfo::run_serial() {
	using namespace YAML;
	InFileINTs intfile(input_file_);
	LOG("generating loopedia file '" << loopedia_file_ << "'");
	ofstream loopedia(loopedia_file_.c_str());
	if (!loopedia)
		ERROR("can't open output file `" << loopedia_file_ << "'");
	set<INT> ints;
	intfile.get_all(ints);
	intfile.close();
	map<Sector, int> numints;
	for (set<INT>::const_iterator i = ints.begin(); i != ints.end(); ++i)
		++numints[i->get_sector()];
	LOG("found " << ints.size() << " integrals from " << numints.size()
			<< " sectors in file '" << input_file_ << "'");
	YAML::Emitter ye;
	ye << BeginMap << Key << "integrals" << Value;
	ye << BeginSeq;
	for (set<INT>::const_iterator i = ints.begin(); i != ints.end(); ++i) {
		LOG("integral " << *i);
		stringstream ss;
		i->to_form_stream(ss);

		Graph graph;
		bool det_non_zero;
		if (!i->get_sector().find_graph(graph, det_non_zero, false))
			ERROR("can't construct graph for " << i->get_sector());
		//graph.print_dot("bla/" + safe_filename(ss.str())+".dot");
		//set<int> bla;
		Topology vacuum = graph;
		LOG("  analyzing biconnected components of vacuum topology");
		vacuum.remove_external_legs();
		if (vacuum.biconnected_components().size() != 1) {
			LOG("  is factorizable, skip it");
			continue;
		}

		LOG("  merging multiple legs entering same vertex");
		graph.merge_external_legs();
		//graph.print_dot("bla/" + safe_filename(ss.str())+"merged.dot");

		LOG("  generating YAML output");
		ye << BeginMap;

		ye << Key << "name" << DoubleQuoted << Value << ss.str();

		ye << Key << "adjacency_list" << Value << Flow << BeginSeq;

		list<Edge> eedges = graph.external_edges();
		list<Edge> iedges = graph.internal_edges();
		list<Edge> aedges;
		list<Edge>::const_iterator e;
		for (e = eedges.begin(); e != eedges.end() ; ++e) {
			ye << Flow << BeginSeq << e->from << e->to << EndSeq;
			aedges.push_back(*e);
		}
		for (e = iedges.begin(); e != iedges.end() ; ++e) {
			ye << Flow << BeginSeq << Flow << e->from << e->to << EndSeq;
			aedges.push_back(*e);
		}
		ye << EndSeq;

		ye << Key << "num_loops" << Value << graph.find_cycles().size();

		ye << Key << "num_edges" << Value << graph.external_edges().size();

		ye << Key << "mass_squares" << Value << Flow << BeginSeq;
		map<GiNaC::ex, int, GiNaC::ex_is_less> m2i; // relabel different mass squares to 1,2,..
		m2i[0] = 0;
		for (e = aedges.begin(); e != aedges.end() ; ++e) {
			GiNaC::ex m2 = graph.get_squaredmass(e->id).expand();
			int nextindex = m2i.size();
			if (m2i.find(m2) == m2i.end())
				m2i[m2] = nextindex;
			ye << m2i[m2];
		}
		ye << EndSeq;

		ye << Key << "num_integrals_in_sector" << Value << numints[i->get_sector()];

		ye << EndMap << Newline;
	}
	ye << EndSeq << EndMap;
    loopedia << ye.c_str() << '\n';
	loopedia.close();
}

}
