/*  topology.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "topology.h"

#include <fstream>
#include <sstream>
#include <cstdlib>
#include "functions.h"
#include "files.h"
#include "yamlutils.h"
#include "combinatorics.h"

using namespace std;
using GiNaC::ex;
using GiNaC::symbol;
using GiNaC::lst;

namespace Reduze {

bool mass_is_less::operator()(const GiNaC::ex& e1, const GiNaC::ex& e2) const {
	using namespace GiNaC;
	if (is_a<numeric>(e1) && is_a<numeric>(e2))
		return ex_to<numeric>(e1) < ex_to<numeric>(e2);
	if (is_a<numeric>(e1) && !is_a<numeric>(e2))
		return true;
	if (!is_a<numeric>(e1) && is_a<numeric>(e2))
		return false;
	ostringstream s1, s2;
	s1 << e1;
	s2 << e2;
	return s1.str() < s2.str();
}

// canonical label

int CanonicalLabel::compare(const CanonicalLabel& other) const {

	// compare propagator types lexicographically
	if (propagator_types_ != other.propagator_types_)
		return ((propagator_types_ < other.propagator_types_) ? -1 : 1);

	// compare masses lexicographically
	int m1 = masses_.nops();
	int m2 = other.masses_.nops();
	lst::const_iterator l1 = masses_.begin(), l2 = other.masses_.begin();
	for (; l1 != masses_.end() && l2 != other.masses_.end(); ++l1, ++l2)
		if (!l1->is_equal(*l2)) {
			mass_is_less mil;
			return ((mil(*l1, *l2)) ? -1 : 1);
		}
	if (m1 != m2)
		return m1 - m2;

	if (adjacencies_ != other.adjacencies_)
		return (adjacencies_ < other.adjacencies_) ? -1 : 1;

	if (node_colorings_ != other.node_colorings_)
		return ((lexicographical_compare(node_colorings_.begin(),
				node_colorings_.end(), other.node_colorings_.begin(),
				other.node_colorings_.end())) ? -1 : 1);
	return 0;
}

bool CanonicalLabel::operator<(const CanonicalLabel& other) const {
	return (this->compare(other) < 0);
}

bool CanonicalLabel::operator==(const CanonicalLabel& other) const {
	return (this->compare(other) == 0);
}

bool CanonicalLabel::operator!=(const CanonicalLabel& other) const {
	return !(*this == other);
}

std::ostream& operator<<(std::ostream& os, const CanonicalLabel& label) {
	YAML::Emitter ye;
	ye << label;
	os << ye.c_str();
	return os;
}

YAML::Emitter& operator<<(YAML::Emitter& ye, const CanonicalLabel& label) {
	using namespace YAML;
	ye << BeginMap;
	ye << Key << "propagator_types" << Value << Flow << label.propagator_types_;
	ye << Key << "masses" << Value << Flow << label.masses_;
	ye << Key << "adjacencies" << Value << BeginSeq;
	for (unsigned i = 0; i < label.adjacencies_.size(); ++i)
		ye << Flow << label.adjacencies_[i];
	ye << EndSeq;
	ye << Key << "node_colorings" << Value << Flow << label.node_colorings_;
	ye << EndMap;
	return ye;
}

void CanonicalLabel::read(const YAML::Node& yn, const GiNaC::lst& symbs) {
	using namespace YAML;
	if (yn.Type() != NodeType::Map || yn.size() != 4)
		throw runtime_error("expected a 4-element map " + position_info(yn));

	const Node& prop_node = yn["propagator_types"];
	if (prop_node.Type() != NodeType::Sequence)
		throw runtime_error("expected a sequence " + position_info(prop_node));
	list<int> props;
	prop_node >> props;
	propagator_types_.swap(props);

	const Node& mass_node = yn["masses"];
	GiNaC::lst masses;
	if (mass_node.Type() != NodeType::Sequence)
		throw runtime_error("expected a sequence " + position_info(mass_node));
	Reduze::read<GiNaC::ex>(mass_node, masses, symbs);
	masses_ = masses;
	unsigned num_colors = masses_.nops();

	const Node& ad_node = yn["adjacencies"];
	if (ad_node.Type() != NodeType::Sequence)
		throw runtime_error("expected a sequence " + position_info(ad_node));
	vector<map<int, colored_multi_edge> > ad;
	ad.reserve(ad_node.size());
	for (Iterator i = ad_node.begin(); i != ad_node.end(); ++i) {
		ad.push_back(map<int, colored_multi_edge>());
		const Node& map_node = *i;
		if (map_node.Type() != NodeType::Map)
			throw runtime_error("expected a map " + position_info(map_node));
		for (Iterator j = map_node.begin(); j != map_node.end(); ++j) {
			int key;
			j.first() >> key;
			colored_multi_edge cme(num_colors);
			const Node& cme_node = j.second();
			cme_node >> cme;
			if (!ad.back().insert(make_pair(key, cme)).second)
				throw runtime_error("node " + position_info(j.first()) + " has already been read.");
		}
	}
	adjacencies_.swap(ad);

	const Node& col_node = yn["node_colorings"];
	if (col_node.Type() != NodeType::Sequence
			|| (col_node.size() != 0 && col_node.size() != adjacencies_.size()))
		throw runtime_error(
				"expected an empty or " + to_string(adjacencies_.size())
						+ "-element sequence " + position_info(col_node));
	vector<int> col;
	col_node >> col;
	node_colorings_.swap(col);
}

// CanonicalRelabeling

std::map<int, int> CanonicalRelabeling::find_node_permutation(
		const std::vector<int>& from, const std::vector<int>& to) {
	VERIFY(from.size() == to.size());
	std::map<int, int> res;
	for (size_t n = 0; n < from.size(); ++n)
		res[from[n]] = to[n];
	return res;
}

std::map<int, int> //
CanonicalRelabeling::find_node_permutation(const CanonicalRelabeling& to) const {
	return CanonicalRelabeling::find_node_permutation(node_permutation_,
			to.node_permutation_);
}
const std::vector<int>& CanonicalRelabeling::node_permutation() const {
	return node_permutation_;
}

YAML::Emitter& operator<<(YAML::Emitter& ye, const CanonicalRelabeling& relabel) {
	ye << YAML::Flow << relabel.node_permutation_;
	return ye;
}
void operator>>(const YAML::Node& n, CanonicalRelabeling& relabel) {
	if (n.Type() != YAML::NodeType::Sequence)
		throw runtime_error("expected a sequence " + position_info(n));
	n >> relabel.node_permutation_;
}
std::ostream& operator<<(std::ostream& os, const CanonicalRelabeling& relabel) {
	YAML::Emitter ye;
	ye << relabel;
	os << ye.c_str();
	return os;
}

// Edge

int Edge::opposite(int node) const {
	if (node == from)
		return to;
	else if (node == to)
		return from;
	else
		ABORT("node " << node << " not connected by this edge: " << *this);
	return 0; // suppress compiler warning
}

void Edge::reverse() {
	swap(from, to);
}

bool Edge::is_self_loop() const {
	return from == to;
}

bool Edge::is_between(int a, int b) const {
	return (a == from && b == to) || (a == to && b == from);
}

int Edge::compare_undirected(const Edge& other) const {
	int amin = std::min(from, to);
	int amax = std::max(from, to);
	int bmin = std::min(other.from, other.to);
	int bmax = std::max(other.from, other.to);
	if (amin != bmin)
		return (amin < bmin ? -1 : 1);
	else if (amax != bmax)
		return (amax < bmax ? -1 : 1);
	else
		return 0;
}

bool Edge::operator<(const Edge& other) const {
	ASSERT(id != other.id || (from == other.from && to == other.to));
	return id < other.id;
}

bool Edge::operator==(const Edge& other) const {
	ASSERT(id != other.id || (from == other.from && to == other.to));
	return id == other.id;
}

std::ostream& operator<<(std::ostream& os, const Edge& e) {
	return os << "(" << e.from << "," << e.to << ";" << e.id << ")";
}

// construction and destruction

Topology::Topology() {
}

Topology::~Topology() {
}

// general

void Topology::set_name(const string& name) {
	name_ = name;
}

std::string Topology::name() const {
	return name_;
}

void Topology::swap(Topology& other, bool call_back) {
	name_.swap(other.name_);
	edges_.swap(other.edges_);
	nodes_.swap(other.nodes_);
	edges_of_node_.swap(other.edges_of_node_);
	if (call_back) {
		swapped();
		other.swapped();
	}
}

// edge and node basics

int Topology::next_edge_id() const {
	return (edges_.empty() ? 1 : (--edges_.end())->first + 1);
}

int Topology::insert_edge(const Edge& e) {
	if (edges_.find(e.id) != edges_.end())
		ABORT("edge " << e.id << " exists already");
	edges_.insert(make_pair(e.id, e));
	if (nodes_.insert(e.from).second)
		inserted_node(e.from);
	if (nodes_.insert(e.to).second)
		inserted_node(e.to);
	edges_of_node_[e.from].push_back(e);
	if (e.to != e.from)
		edges_of_node_[e.to].push_back(e);
	inserted_edge(e.id);
	return e.id;
}

int Topology::insert_edge(int from, int to) {
	return insert_edge(Edge(from, to, next_edge_id()));
}

void Topology::remove_edge(int edge_id) {
	map<int, Edge>::iterator eit = edges_.find(edge_id);
	if (eit == edges_.end())
		ABORT("no such edge with id " << edge_id << " to remove");
	Edge edge = eit->second;
	int node = edge.from;
	while (true) {
		list<Edge>& edges = edges_of_node_[node];
		for (list<Edge>::iterator e = edges.begin(); e != edges.end();)
			if (e->id == edge_id)
				edges.erase(e++);
			else
				++e;
		if (node == edge.to)
			break;
		node = edge.to;
	}
	edges_.erase(edge_id);
	removed_edge(edge_id);
}

void Topology::contract_edge(int edge_id, bool remove_greater_node) {
	map<int, Edge>::iterator eit = edges_.find(edge_id);
	if (eit == edges_.end())
		ABORT("no such edge with id " << edge_id << " to contract");
	const Edge& e = eit->second;
	if (remove_greater_node)
		identify_nodes(std::max(e.from, e.to), std::min(e.from, e.to));
	else
		identify_nodes(std::min(e.from, e.to), std::max(e.from, e.to));
	remove_edge(edge_id);
}

void Topology::contract_bridges() {
	list<Topology> bcc = biconnected_components();
	for (list<Topology>::const_iterator c = bcc.begin(); c != bcc.end(); ++c)
		if (c->num_edges() == 1 && //
				!is_external_edge(c->edges().begin()->first) && //
				!c->edges().begin()->second.is_self_loop())
			contract_edge(c->edges().begin()->first);
}

void Topology::remove_node(int node) {
	VERIFY(!nodes().empty());
	int last_id = *nodes().rbegin();
	map<int, int> newnodes; // new node id to use for a given edge
	map<int, int>::const_iterator nn;

	// update edges for 'node', edges by id
	list<Edge>& es = edges_of_node_[node];
	for (list<Edge>::const_iterator e = es.begin(); e != es.end(); ++e) {
		Edge& ee = edges_[e->id];
		int newnode = ++last_id;
		nodes_.insert(newnode);
		ee.from = (ee.from == node ? newnode : ee.from);
		ee.to = (ee.to == node ? newnode : ee.to);
		newnodes[ee.id] = newnode;
		edges_of_node_[newnode].push_back(ee);
	}

	// update edges of neighbor nodes
	set<int> nbnodes = neighbors(node);
	for (set<int>::const_iterator n = nbnodes.begin(); n != nbnodes.end(); ++n) {
		list<Edge>& es = edges_of_node_[*n];
		for (list<Edge>::iterator e = es.begin(); e != es.end(); ++e) {
			if ((nn = newnodes.find(e->id)) != newnodes.end()) {
				e->from = (e->from == node ? nn->second : e->from);
				e->to = (e->to == node ? nn->second : e->to);
			}
		}
	}

	// signal insertion of new nodes
	for (nn = newnodes.begin(); nn != newnodes.end(); ++nn)
		inserted_node(nn->second);

	nodes_.erase(node);
	removed_node(node);
}

void Topology::remove_isolated_node(int node) {
	const list<Edge>& es = edges_of_node(node);
	if (!es.empty())
		ABORT("Node " << node << " is not isolated.");
	nodes_.erase(node);
	edges_of_node_.erase(node);
}

void Topology::identify_nodes(int from, int to) {
	if (from == to)
		return;

	// update edges for 'to', edges by id
	list<Edge>& es = edges_of_node_[from];
	for (list<Edge>::const_iterator e = es.begin(); e != es.end(); ++e) {
		Edge& ee = edges_[e->id];
		bool new_adj = !(ee.from == to || ee.to == to);
		ee.from = (ee.from == from ? to : ee.from);
		ee.to = (ee.to == from ? to : ee.to);
		if (new_adj)
			edges_of_node_[to].push_back(ee);
	}

	// update edges of neighbor nodes
	set<int> nnodes = neighbors(from);
	for (set<int>::const_iterator n = nnodes.begin(); n != nnodes.end(); ++n) {
		list<Edge>& es = edges_of_node_[*n];
		for (list<Edge>::iterator e = es.begin(); e != es.end(); ++e) {
			e->from = (e->from == from ? to : e->from);
			e->to = (e->to == from ? to : e->to);
		}
	}

	edges_of_node_.erase(from);
	nodes_.erase(from);

	removed_node(from);
}

void Topology::reverse_edge(int edge_id) {
	Edge& e = edges_[edge_id];
	e.reverse();
	list<Edge>& f = edges_of_node_[e.from];
	for (list<Edge>::iterator e = f.begin(); e != f.end(); ++e)
		if (e->id == edge_id) {
			e->reverse();
			break;
		}
	list<Edge>& t = edges_of_node_[e.to];
	for (list<Edge>::iterator e = t.begin(); e != t.end(); ++e)
		if (e->id == edge_id) {
			e->reverse();
			break;
		}
}

void Topology::remove_external_legs() {
	const list<Edge> ext = external_edges();
	for (list<Edge>::const_iterator e = ext.begin(); e != ext.end(); ++e) {
		bool from_is_external = is_external_node(e->from);
		bool to_is_external = is_external_node(e->to);
		remove_edge(e->id);
		if (from_is_external)
			remove_isolated_node(e->from);
		if (to_is_external)
			remove_isolated_node(e->to);
	}
}

void Topology::join_external_nodes() {
	const list<Edge> ext = external_edges();
	if (ext.empty())
		return;
	ASSERT(!nodes().empty());
	int n = std::max(1, *nodes().rbegin() + 1);
	for (list<Edge>::const_iterator e = ext.begin(); e != ext.end(); ++e) {
		Edge edge = *e; // copy
		if (is_external_node(e->from))
			edge.from = n;
		if (is_external_node(e->to))
			edge.to = n;
		remove_edge(e->id);
		if (edge.from != e->from)
			remove_isolated_node(e->from);
		if (edge.to != e->to)
			remove_isolated_node(e->to);
		insert_edge(edge);
	}
}

void Topology::insert_node(int node_id) {
	nodes_.insert(node_id);
	edges_of_node_[node_id];
	inserted_node(node_id);
}

void Topology::inserted_node(int) {
}

void Topology::inserted_edge(int) {
}

void Topology::removed_node(int) {
}

void Topology::removed_edge(int) {
}

void Topology::swapped() {
}

std::pair<unsigned int, unsigned int> Topology::cleave(int node,
		const std::set<Edge>& edges, int sign, int new_edge) {
	if (!contains_node(node))
		ABORT("Topology " << name_ << " doesn't contain node " << node);
	//int new_node = max(1, *nodes_.rbegin() + 1);
	// gives a nicer numbering
	int new_node = 1;
	while (nodes_.find(new_node) != nodes_.end())
		++new_node;

	list<Edge>& es = edges_of_node_[node];
	unsigned int count = 0;
	for (list<Edge>::iterator e = es.begin(); e != es.end();)
		if (edges.find(*e) != edges.end())
			es.erase(e++), ++count;
		else
			++e;

	if (count != edges.size())
		ABORT("Mismatch in given edges and edges connecting node " << node);

	if (sign > 0)
		insert_edge(Edge(new_node, node, new_edge));
	else
		insert_edge(Edge(node, new_node, new_edge));

	// update edges
	map<int, Edge>::iterator ma;
	for (set<Edge>::const_iterator m = edges.begin(); m != edges.end(); ++m) {
		if ((ma = edges_.find(m->id)) != edges_.end()) {
			if (ma->second.from == node)
				ma->second.from = new_node;
			else if (ma->second.to == node)
				ma->second.to = new_node;
			else
				ABORT("edge " << ma->first << " doesn't connect to node " << node);
		} else {
			ABORT("edge " << *m << " not contained in edges of topology " << name_);
		}
		edges_of_node_[new_node].push_back(*m);
	}

	// update edges of node
	for (map<int, list<Edge> >::iterator it = edges_of_node_.begin();
			it != edges_of_node_.end(); ++it) {
		for (list<Edge>::iterator e = it->second.begin(); e != it->second.end(); ++e) {
			if (edges.find(*e) != edges.end()) {
				if (e->from == node)
					e->from = new_node;
				if (e->to == node)
					e->to = new_node;
			}
		}
	}

	return make_pair(new_node, new_edge);
}

void Topology::remove_degree_2_nodes() {
	set<int> removed_edges;
	set<int> nodes = nodes_; // copy
	// reverse iteration to delete nodes with larger id first
	for (set<int>::const_reverse_iterator n = nodes.rbegin(); n != nodes.rend();
			++n) {
		if (!contains_node(*n))
			continue;
		const list<Edge>& edges = edges_of_node(*n);
		if (edges.size() != 2)
			continue;
		// e1 --- n --- e2  :  delete node n and edge with greater id
		Edge e1 = *edges.begin();
		Edge e2 = *edges.rbegin();
		if (e1.is_self_loop() || e2.is_self_loop())
			continue;
		if (is_external_edge(e1.id) && is_external_edge(e2.id)) {
			WARNING("Node " << *n << " with two external edges attached not removed.");
			continue;
		}
		Edge del = (e1.id < e2.id ? e2 : e1);
		ASSERT(*n == del.from || *n == del.to);
		bool delete_larger_node = (*n == std::max(del.from, del.to));
		contract_edge(del.id, delete_larger_node);
	}
}

Edge Topology::insert_degree_2_node(int edge_id) {
	Edge e = edge(edge_id);
	int node = std::max(1, *nodes_.rbegin() + 1);
	bool is_outgoing = is_external_node(e.to);
	remove_edge(edge_id);
	// keep original edge id on external legs
	if (is_outgoing) {
		int from = e.from;
		e.from = node;
		insert_edge(e);
		int id = insert_edge(from, node);
		return Edge(from, node, id);
	}
	int to = e.to;
	e.to = node;
	insert_edge(e);
	int id = insert_edge(node, to);
	return Edge(node, to, id);
}

// getters

unsigned int Topology::num_edges() const {
	return edges_.size();
}

unsigned int Topology::num_nodes() const {
	return nodes_.size();
}

unsigned int Topology::num_external_nodes() const {
	unsigned int count = 0;
	for (set<int>::const_iterator n = nodes_.begin(); n != nodes_.end(); ++n)
		if (is_external_node(*n))
			++count;
	return count;
}

unsigned int Topology::num_internal_edges() const {
	unsigned int count = 0;
	map<int, Edge>::const_iterator e;
	for (e = edges_.begin(); e != edges_.end(); ++e)
		if (!is_external_edge(e->first))
			++count;
	return count;
}

bool Topology::contains_node(int node) const {
	return nodes_.find(node) != nodes_.end();
}

bool Topology::contains_edge(int edge) const {
	return edges_.find(edge) != edges_.end();
}

bool Topology::is_external_edge(int edge_id) const {
	map<int, Edge>::const_iterator it = edges_.find(edge_id);
	if (it == edges_.end())
		ABORT("Edge " << edge_id << " does not exist");
	return (is_external_node(it->second.from)
					|| is_external_node(it->second.to));
}

bool Topology::is_external_node(int node) const {
	const list<Edge>& es = edges_of_node(node);
	ASSERT(nodes_.find(node) != nodes_.end());
	if (es.size() != 1)
		return false;
	return !es.begin()->is_self_loop();
}

const std::set<int>& Topology::nodes() const {
	return nodes_;
}

const std::map<int, Edge>& Topology::edges() const {
	return edges_;
}

const list<Edge>& Topology::edges_of_node(int node) const {
	map<int, list<Edge> >::const_iterator i = edges_of_node_.find(node);
	if (i == edges_of_node_.end())
		ABORT("no such node " << node);
	return i->second;
}

const map<int, list<Edge> >& Topology::edges_of_nodes() const {
	return edges_of_node_;
}

const Edge& Topology::edge(int edge_id) const {
	map<int, Edge>::const_iterator it = edges_.find(edge_id);
	if (it == edges_.end())
		ABORT("Edge " << edge_id << " does not exist");
	return it->second;
}

std::list<Edge> Topology::find_edges_between(int node1, int node2) const {
	list<Edge> res;
	for (map<int, Edge>::const_iterator e = edges_.begin(); e != edges_.end(); ++e)
		if (e->second.is_between(node1, node2))
			res.push_back(e->second);
	return res;
}

std::set<int> Topology::neighbors(int node) const {
	set<int> neighbors;
	list<Edge> es = edges_of_node(node);
	for (list<Edge>::const_iterator e = es.begin(); e != es.end(); ++e) {
		neighbors.insert(e->from);
		neighbors.insert(e->to);
	}
	neighbors.erase(node);
	return neighbors;
}

std::set<int> Topology::internal_nodes() const {
	set<int> res;
	set<int>::iterator pos = res.begin();
	for (set<int>::const_iterator n = nodes_.begin(); n != nodes_.end(); ++n)
		if (!is_external_node(*n))
			pos = res.insert(pos, *n);
	return res;
}

std::set<int> Topology::external_nodes() const {
	set<int> res;
	set<int>::iterator pos = res.begin();
	for (set<int>::const_iterator n = nodes_.begin(); n != nodes_.end(); ++n)
		if (is_external_node(*n))
			pos = res.insert(pos, *n);
	return res;
}

std::list<Edge> Topology::internal_edges() const {
	list<Edge> result;
	map<int, Edge>::const_iterator e;
	for (e = edges_.begin(); e != edges_.end(); ++e)
		if (!is_external_edge(e->first))
			result.push_back(e->second);
	return result;
}

std::list<Edge> Topology::external_edges() const {
	list<Edge> result;
	map<int, Edge>::const_iterator e;
	for (e = edges_.begin(); e != edges_.end(); ++e)
		if (is_external_edge(e->first))
			result.push_back(e->second);
	return result;
}

std::set<int> Topology::find_loop_edges() const {
	set<int> inter;
	one_pi_components(inter);
	set<int> result;
	map<int, Edge>::const_iterator e;
	for (e = edges_.begin(); e != edges_.end(); ++e)
		if (!is_external_edge(e->first) && inter.find(e->first) == inter.end())
			result.insert(e->first);
	return result;
}

std::pair<std::set<int>, std::set<int> > Topology::spanning_tree_and_cycles(
		int node) const {
	set<int> tree, cycles; // edges: spanning tree, cycle
	set<int> done, todo; // nodes: visited, yet to visit

	for (todo.insert(node); !todo.empty();) {
		int n = *todo.begin();
		list<Edge> edges = edges_of_node(n);
		for (list<Edge>::const_iterator e = edges.begin(); e != edges.end();
				++e) {
			if (tree.find(e->id) != tree.end()
					|| cycles.find(e->id) != cycles.end())
				continue;
			int next = e->opposite(n);
			if (done.find(next) == done.end()
					&& todo.find(next) == todo.end()) {
				todo.insert(next);
				tree.insert(e->id);
			} else {
				cycles.insert(e->id);
			}
		}
		todo.erase(n);
		done.insert(n);
	}
	return make_pair(tree, cycles);
}

std::set<int> Topology::find_cycles() const {
	set<int> cycles;
	const list<Topology> comp = connected_components();
	for (list<Topology>::const_iterator t = comp.begin(); t != comp.end();
			++t) {
		ASSERT(!t->nodes_.empty());
		int node = *t->nodes_.begin();
		pair<set<int>, set<int> > tmp = t->spanning_tree_and_cycles(node);
		cycles.insert(tmp.second.begin(), tmp.second.end());
	}
	// compare size of cycles with alternative calculation of first Betti number
	unsigned int nc = num_edges() - num_nodes() + comp.size();

	/*cout << "topo   " << name_ << endl;
	 cout << "cycles " << cycles.size() << "\n";
	 cout << "edges  " << num_edges() << "\n";
	 cout << "nodes  " << num_nodes() << "\n";
	 cout << "cc     " << connected_components().size() << "\n";
	 cout << "nc     " << nc << "\n";
	 */
	ASSERT(cycles.size() == nc);
	return cycles;
}

// topological analysis

bool Topology::is_connected() const {
	return connected_components().size() == 1;
}

bool Topology::is_one_particle_irreducible() const {
	set<int> edges;
	return one_pi_components(edges).size() == 1;
}

Topology Topology::connected_component(int node, int disabled_edge) const {
	Topology t;
	if (!contains_node(node))
		return t;
	t.insert_node(node); // for special case of node without edges
	set<int> todo_nodes, visited_nodes;
	for (todo_nodes.insert(node); !todo_nodes.empty();) {
		int n = *todo_nodes.begin();
		list<Edge> edges = edges_of_node(n);
		for (list<Edge>::const_iterator e = edges.begin(); e != edges.end();
				++e) {
			if (e->id == disabled_edge)
				continue;
			int next = e->opposite(n);
			if (visited_nodes.find(next) == visited_nodes.end())
				todo_nodes.insert(next);
			if (!t.contains_edge(e->id))
				t.insert_edge(*e);
		}
		todo_nodes.erase(n);
		visited_nodes.insert(n);
	}
	return t;
}

list<Topology> Topology::connected_components() const {
	list<Topology> comps;
	set<int> nodes = nodes_;
	while (!nodes.empty()) {
		int node = *nodes.begin();
		nodes.erase(nodes.begin());
		Topology t = connected_component(node);
		set<int>::const_iterator it;
		for (it = t.nodes_.begin(); it != t.nodes_.end(); ++it)
			nodes.erase(*it);
		comps.push_back(t);
	}
	return comps;
}

/*
 string indent(int n) {
 string s;
 for (int i = 0; i < n; ++i)
 s += ' ';
 return s;
 }
 */

// Implementation might be included in DFSPalmTree some day. For now, we keep
// it separate, since DFSPalmTree works only for restricted edge/node numbers.
// NOTE: unlike in the original algorithm, we use original node ids in lowpt
// (and not the positions in DFS !)
void find_biconnected(const Topology& topo, //
		int v, // current node
		set<int>& todo, // unvisited nodes (separate variable for disconn. comps)
		map<int, int>& pos, // position wrt DFS search
		map<int, int>& lowpt, // node with lowest position reachable by second path
		map<int, bool>& is_arc, // true if edge is tree arc, false if frond
		stack<Edge>& edges, list<Topology>& comps, // return biconnected components
		set<int>* cuts // return cut nodes if non-zero
		) {
	todo.erase(v);
	int num_numbers = pos.size();
	pos[v] = num_numbers + 1;
	lowpt[v] = v;

	//cout << indent(pos[v]) << "|" << v << "| visit node\n";
	int num_chops = 0; // number of components chopped off current node
	const list<Edge>& el = topo.edges_of_node(v);
	for (list<Edge>::const_iterator e = el.begin(); e != el.end(); ++e) {
		if (is_arc.find(e->id) != is_arc.end()) // need one visit only
			continue;
		//cout << indent(pos[v]) << "visiting edge " << *e << "\n";
		edges.push(*e);
		int w = e->opposite(v); // son of v
		if (pos.find(w) == pos.end()) { // w is new node (edge: tree arc)
			is_arc[e->id] = true; // TreeArc
			find_biconnected(topo, w, todo, pos, lowpt, is_arc, edges, comps,
					cuts);
			lowpt[v] = (pos[lowpt[w]] < pos[lowpt[v]] ? lowpt[w] : lowpt[v]);
		} else { // w is old node (edge: palm frond)
			is_arc[e->id] = false; // PalmFrond
			lowpt[v] = (pos[w] < pos[lowpt[v]] ? w : lowpt[v]);
		}
		if ((w == v) // self-loop
		|| (is_arc[e->id] && lowpt[w] == w) // single edge
				|| (is_arc[e->id] && lowpt[w] == v) // loop bcc
				) {
			// found cut node or root_node with last component
			bool is_root_node = (pos[v] == 1);
			if (cuts != 0 && !(is_root_node && num_chops == 0))
				cuts->insert(v); // is really a cut node
			// cout << indent(pos[v]) << "CHOP OFF COMPONENT ! at " << v << endl;
			// chop off a component
			Topology comp;
			while (true) {
				ASSERT(!edges.empty());
				Edge e = edges.top();
				edges.pop();
				//cout << indent(pos[v]) << "inserting " << e << endl;
				comp.insert_edge(e);
				if (e.from == v && e.to == v) // self-loop (frond)
					break;
				if ((e.from == v || e.to == v) && is_arc[e.id]) // first arc
					break;
			}
			++num_chops;
			comps.push_back(comp);
		}
	}
	//cout << indent(pos[v]) << "done with " << v << "\n";
}

list<Topology> Topology::biconnected_components(set<int>* cut_nodes) const {
	list<Topology> bcc;
	set<int> todo = nodes_;
	while (!todo.empty()) {
		map<int, int> pos, lowpt;
		map<int, bool> edge_type;
		stack<Edge> edges;
		find_biconnected(*this, *todo.begin(), todo, pos, lowpt, edge_type,
				edges, bcc, cut_nodes);
	}
	/*
	 LOGXX("found " << bcc.size() << " biconnected components");
	 int count = 0;
	 for (list<Topology>::iterator b = bcc.begin(); b != bcc.end(); ++b, ++count)
	 b->print_postscript("bcc" + to_string(count) + ".ps");
	 */
	return bcc;
}

list<Topology> Topology::biconnected_vacuum_components(
		list<Topology>& other) const {
	set<int> cutset;
	list<Topology> bcclist = biconnected_components(&cutset);
	LOGXX("found " << bcclist.size() << " biconnected components");
	vector<Topology> bcc(bcclist.begin(), bcclist.end()); // biconnect. comps
	vector<set<int> > cuts(bcc.size()); // connecting nodes for each comp
	map<int, set<int> > comps; // comps for each cut node
	list<int> vtodo; // components identified as vacuum but not processed yet
	// find cut and external nodes of biconnected components
	for (size_t c = 0; c < bcc.size(); ++c) {
		const set<int>& ns = bcc[c].nodes();
		for (set<int>::const_iterator n = ns.begin(); n != ns.end(); ++n)
			if (is_external_node(*n)) {
				cuts[c].insert(*n); // like cut node (never removed)
			} else if (cutset.find(*n) != cutset.end()) {
				cuts[c].insert(*n);
				comps[*n].insert(c);
			}
		if (cuts[c].size() <= 1)
			vtodo.push_back(c);
	}

	const map<int, set<int> > comps_backup(comps); // copy

	// ignore vacuum component in other components
	while (!vtodo.empty()) {
		size_t v = vtodo.front();
		vtodo.pop_front();
		if (cuts[v].size() == 1) { // connected comp., update rest
			int n = *cuts[v].begin(); // cut node
			comps[n].erase(v); // ignore v for cut node
			if (comps[n].size() == 1) { // one comp. left at n
				int next = *comps[n].begin();
				cuts[next].erase(n); // ignore n
				if (cuts[next].size() <= 1) // next is vacuum
					vtodo.push_back(next);
			}
		}
	}

	// assume: vacuum components have 0 or 1 cut node

	int lauf = std::max(1, *nodes_.rbegin() + 1);
	map<int, map<int, int> > rename; // nodes to be renamed for each vacuum component
	for (map<int, set<int> >::const_iterator c = comps_backup.begin();
			c != comps_backup.end(); ++c) {
		for (set<int>::const_iterator s = c->second.begin();
				s != c->second.end(); ++s)
			if (cuts[*s].size() <= 1)
				rename[*s].insert(make_pair(c->first, lauf++));
	}

	list<Topology> vacuum;
	for (size_t c = 0; c < bcc.size(); ++c) {
		if (cuts[c].size() > 1) {
			other.push_back(bcc[c]);
		} else {
			vacuum.push_back(Topology());
			Topology& vac = vacuum.back();
			const map<int, Edge>& es = bcc[c].edges();
			const map<int, int>& re = rename[c];
			for (map<int, Edge>::const_iterator e = es.begin(); e != es.end();
					++e) {
				Edge ed = e->second;
				map<int, int>::const_iterator r1 = re.find(ed.from);
				map<int, int>::const_iterator r2 = re.find(ed.to);
				if (r1 != re.end())
					ed.from = r1->second;
				if (r2 != re.end())
					ed.to = r2->second;
				vac.insert_edge(ed);
			}
		}
	}

	LOGXX("found " << vacuum.size() << " biconnected vacuum components");
	return vacuum;
}

list<Topology> Topology::one_pi_components(set<int>& inter_edges) const {
	list<Topology> result;
	if (!is_connected()) { // disconnected
		list<Topology> c = connected_components();
		for (list<Topology>::const_iterator t = c.begin(); t != c.end(); ++t) {
			list<Topology> l = t->one_pi_components(inter_edges);
			result.splice(result.end(), l);
		}
	} else { // connected
		Topology t1;
		map<int, Edge>::const_iterator eit;
		for (eit = edges_.begin(); eit != edges_.end(); ++eit) {
			if (is_external_edge(eit->first))
				continue;
			t1 = connected_component(eit->second.from, eit->first);
			if (!t1.contains_node(eit->second.to)) // found reduction
				break;
		}
		if (eit != edges_.end()) { // one particle reducible
			Topology t2 = connected_component(eit->second.to, eit->first);
			list<Topology> l1 = t1.one_pi_components(inter_edges);
			list<Topology> l2 = t2.one_pi_components(inter_edges);
			inter_edges.insert(eit->first);
			result.splice(result.end(), l1);
			result.splice(result.end(), l2);
		} else { // 1pi
			result.push_back(*this);
		}
	}
	return result;
}

int Topology::num_loop_edges() const {
	set<int> dummy;
	std::list<Topology> comp = one_pi_components(dummy);
	int num = 0;
	for (list<Topology>::const_iterator c = comp.begin(); c != comp.end(); ++c)
		num += c->num_internal_edges();
	return num;
}

// public functions for isomorphisms

std::pair<CanonicalLabel, CanonicalRelabeling> Topology::find_canonical_label(
		bool perm_ext_nodes) const {

	pair<map<int, int>, lst> ecm = get_edge_coloring(); // virtual
	const map<int, int>& color_of_edge = ecm.first;
	const lst& masses = ecm.second;
	UndirectedGraph ungraph(nodes_, edges_, color_of_edge);

	map<int, int> color_of_node = get_node_coloring(perm_ext_nodes);
	VERIFY(color_of_node.empty() || color_of_node.size() == num_nodes());

	bool strict_pruning = true;
	set<int> no_free_permutations;
	pair<vector<map<int, colored_multi_edge> >, vector<int> > ug_label;
	ug_label = ungraph.find_canonical_label(color_of_node, no_free_permutations,
			strict_pruning);

	vector<int> colorings(color_of_node.size());
	for (unsigned int i = 0; i < color_of_node.size(); ++i)
		colorings[i] = color_of_node.at(ug_label.second[i]);

	CanonicalLabel label;
	label.adjacencies_.swap(ug_label.first);
	label.node_colorings_.swap(colorings);
	label.masses_ = masses;
	CanonicalRelabeling relabel;
	relabel.node_permutation_.swap(ug_label.second);
	return make_pair(label, relabel);
}

std::list<std::map<int, int> > Topology::find_node_symmetry_group(
		bool prune_but_not_strictly) const {

	pair<map<int, int>, lst> ecm = get_edge_coloring(); // virtual
	const map<int, int>& color_of_edge = ecm.first;
	//const lst& masses = ecm.second;
	UndirectedGraph ungraph(nodes_, edges_, color_of_edge);

	map<int, int> color_of_node;

	list<map<int, int> > perms;
	set<int> no_free_permutations;
	bool strict_pruning = true;
	if (prune_but_not_strictly) {
		no_free_permutations = external_nodes();
		strict_pruning = false;
	}

	// strict pruning by default set to false to allow for some more permutations of ext. nodes
	ungraph.find_node_symmetry_group(perms, color_of_node, no_free_permutations,
			strict_pruning);

	return perms;
}

void Topology::find_edge_symmetry_group(
		std::list<std::map<int, std::pair<int, int> > >& edge_permutations,
		bool prune_but_not_strictly) const {

	// find the node symmetry of the line graph
	// the nodes of the line graph are the edge ids of *this
	LOGX("  finding edge symmetry group of graph " << name());
	Topology lg = get_line_graph();
	ASSERT(lg.num_nodes() == num_edges());
	const map<int, int> lg_color_of_edge = lg.get_edge_coloring().first;
	UndirectedGraph ungraph(lg.nodes(), lg.edges(), lg_color_of_edge);

	set<int> no_free_permute;
	bool strict_pruning = true;
	if (prune_but_not_strictly) {
		// dont allow external edges to be permuted freely....
		list<Edge> el = external_edges();
		for (list<Edge>::const_iterator e = el.begin(); e != el.end(); ++e)
			no_free_permute.insert(e->id);
		// ...except in cases where some ext. edges are permuted bc of internal permutations, then permute all
		strict_pruning = false;
	}

	// this edge coloring is the line graphs node coloring
	const map<int, int> node_color = get_edge_coloring().first; // virtual
	list<map<int, int> > edge_perm_list;
	ungraph.find_node_symmetry_group(edge_perm_list, node_color,
			no_free_permute, strict_pruning);
	int num_edge_perms = edge_perm_list.size();

	// determine the sign of the edge permutation
	edge_permutations.clear();
	while (!edge_perm_list.empty()) {
		const map<int, int>& edge_perm = edge_perm_list.front();
		map<int, int> node_perm;
		if (reconstruct_node_permutation(node_perm, edge_perm)) {
			map<int, pair<int, int> > edge_perm_sign;
			map<int, int>::const_iterator ep;
			for (ep = edge_perm.begin(); ep != edge_perm.end(); ++ep) {
				int sign = 1;
				if (node_perm.at(edge(ep->first).from) != edge(ep->second).from)
					sign = -1;
				edge_perm_sign[ep->first] = make_pair(sign, ep->second);
			}
			edge_permutations.push_back(map<int, pair<int, int> >());
			edge_permutations.back().swap(edge_perm_sign);
		}
		edge_perm_list.pop_front();
	}
	int num_node_perms = edge_permutations.size();
	LOGX("  found " << num_edge_perms << " edge permutation");
	LOGX("  found " << num_node_perms //
			<< " edge permutations with valid node permutation");
	if (num_edge_perms != num_node_perms)
		LOGX("    mismatch might come from star K_1_3 as subgraph");
}

namespace {
// turn set1 into the intersection of set1 and set2
void intersect_with(std::set<int>& set1, const std::set<int>& set2) {
	set<int> tmp;
	set<int>::const_iterator s;
	for (s = set1.begin(); s != set1.end(); ++s)
		if (set2.find(*s) != set2.end())
			tmp.insert(*s);
	set1.swap(tmp);
}

bool is_a_permutation(const std::map<int, int>& perm) {
	set<int> test_from, test_to;
	map<int, int>::const_iterator it;
	for (it = perm.begin(); it != perm.end(); ++it) {
		test_from.insert(it->first);
		test_to.insert(it->second);
	}
	return test_from == test_to;
}
}

bool Topology::reconstruct_node_permutation(std::map<int, int>& node_permutation,
		const std::map<int, int>& edge_permutation) const {
	/*
	 * For every node N: transform all adjacent edges and check the
	 * intersection of their endpoints.
	 * If the size of the intersection is:
	 *   0: No node permutation possible for node N, return false
	 *   1: A unique node permutation for node N has been found
	 *   2:	All edges of N transform to edges between two endpoints ep1, ep2.
	 *		The node N must be connected to a single opposite node M.
	 *		N and M are mapped to ep1, ep2, respecting the number of neighbors they have.
	 *      If all 4 nodes have the same number of neighbors choose
	 *        permutation N -> min(intersection) and M -> max(intersection)
	 */

	if (!is_a_permutation(edge_permutation))  {
		ABORT("The edge transformation is not a permutation");
	}
	node_permutation.clear();
	map<int, int> node_map;

	set<int>::const_iterator node;
	for (node = nodes_.begin(); node != nodes_.end(); ++node) {
		if (node_map.find(*node) != node_map.end())
			continue;
		const list<Edge>& edges_from = edges_of_node(*node);

		// permutations of isolated nodes cannot be reconstructed
		if (edges_from.empty())
			continue;

		set<int> inter_nodes_to; // intersection of the endpoints of the mapped edges
		set<int> oppnodes; // nodes opposite of node
		list<Edge>::const_iterator e_from;
		for (e_from = edges_from.begin(); e_from != edges_from.end(); ++e_from) {
			const Edge& e_to = edge(edge_permutation.at(e_from->id));
			set<int> endpoints;
			endpoints.insert(e_to.from);
			endpoints.insert(e_to.to);
			if (e_from == edges_from.begin())
				inter_nodes_to.swap(endpoints);
			else
				intersect_with(inter_nodes_to, endpoints);
			// test for case 0:
			if (inter_nodes_to.empty())
				return false;
			oppnodes.insert(e_from->opposite(*node));
		}
		ASSERT(inter_nodes_to.size() == 1 || inter_nodes_to.size() == 2);
		int ep1 = *inter_nodes_to.begin();
		int num_e_ep1 = edges_of_node(ep1).size();
		int num_e_node = edges_from.size();

		// test for case 1:
		if (inter_nodes_to.size() == 1) {
			if (num_e_node != num_e_ep1)
				return false;
			node_map[*node] = ep1;

		// test for case 2:
		} else {
			int ep2 = *inter_nodes_to.rbegin();
			int num_e_ep2 = edges_of_node(ep2).size();
			if (oppnodes.size() != 1 || *node == *oppnodes.begin())
				return false;
			int opnode = *oppnodes.begin();
			int num_e_opnode = edges_of_node(opnode).size();
			map<int, int>::const_iterator m = node_map.find(opnode);
			if (m != node_map.end()) {
				if (m->second != ep1 && m->second != ep2)
					return false;
				node_map[*node] = ((m->second == ep1) ? ep2 : ep1);
			} else if (num_e_node != num_e_opnode) {
				if (num_e_ep1 != num_e_node && num_e_ep1 != num_e_opnode)
					return false;
				if (num_e_ep2 != num_e_node && num_e_ep2 != num_e_opnode)
					return false;
				node_map[*node] = ((num_e_ep1 == num_e_node) ? ep1 : ep2);
				node_map[opnode] = ((num_e_ep1 == num_e_node) ? ep2 : ep1);
			} else { // node and opnode have the same number of neighbours
				if (num_e_ep1 != num_e_node || num_e_ep2 != num_e_node)
					return false;
				node_map[*node] = ep1;
				node_map[opnode] = ep2;
			}
		}
	}
	if (!is_a_permutation(node_map))  {
		ABORT("The node transformation is not a permutation");
	}
	node_map.swap(node_permutation);
	return true;
}

Topology Topology::get_line_graph() const {
	Topology top;
	top.set_name("line_graph_" + name());
	const map<int, Edge>& t_edges = edges();
	map<int, Edge>::const_iterator mit1, mit2;
	for (mit1 = t_edges.begin(); mit1 != t_edges.end(); ++mit1) {
		const Edge& e1 = mit1->second;
		if (e1.is_self_loop()) {
			// self-loops become self-loops
			top.insert_edge(e1.id, e1.id);
			continue;
		}
		// edges not connected to any other edge become a single node
		top.insert_node(e1.id);
		mit2 = mit1;
		for (++mit2; mit2 != t_edges.end(); ++mit2) {
			const Edge& e2 = mit2->second;
			if (e2.is_self_loop())
				continue;
			if (e1.from == e2.from || e1.from == e2.to)
				top.insert_edge(e1.id, e2.id);
			if (e1.to == e2.from || e1.to == e2.to)
				top.insert_edge(e1.id, e2.id);
		}
	}
	return top;
}

std::pair<std::map<int, int>, GiNaC::lst> Topology::get_edge_coloring() const {
	GiNaC::lst masses;
	masses.append(0);
	map<int, int> color_of_edge;
	for (map<int, Edge>::const_iterator e = edges_.begin(); e != edges_.end(); ++e)
		color_of_edge[e->first] = 0;
	return make_pair(color_of_edge, masses);
}

std::map<int, int> Topology::get_node_coloring(
		bool permute_external_nodes) const {

	map<int, int> color_of_node;
	if (!permute_external_nodes) {
		int lauf = -num_external_nodes();
		set<int>::const_iterator n;
		for (n = nodes_.begin(); n != nodes_.end(); ++n) {
			if (is_external_node(*n))
				color_of_node[*n] = lauf++;
			else
				color_of_node[*n] = 0;
		}
	}
	return color_of_node;
}

// output

void Topology::print_dot(const string& filename) const {
	ofstream of(filename.c_str());
	of << "graph " << (name_ != "graph" ? name_ : "graph0") << " {\n";
	map<int, Edge>::const_iterator it;
	for (it = edges_.begin(); it != edges_.end(); ++it) {
		const Edge& e = it->second;
		of << e.from << " -- " << e.to << " [label=\"" << it->first
				<< "\",fontsize=10];\n";
	}
	// nodes: just the label
	for (set<int>::const_iterator n = nodes_.begin(); n != nodes_.end(); ++n)
		of << *n << " [fontcolor=\"blue\",color=blue];\n";

	// problem with labels for multiple edges using neato:
	//		of << *n
	//				<< " [fontcolor=\"blue\",shape=circle,style=filled,color=blue,fixedsize=true,width=0,height=0];\n";
	//,shape=circle,width=0.15,height=0.15,color=\"blue\""
	of << "}\n";
	of.close();
}

void Topology::print_postscript(const string& filename) const {
	string dotfilename = filename + ".dot";
	print_dot(dotfilename);
	stringstream ss;
	ss << "neato -Tps " << dotfilename << " -o " << filename;
	int ret = system(ss.str().c_str());
	VERIFY(ret == 0);
}

std::ostream& operator<<(std::ostream& os, const Topology& t) {
	os << "Topology: " << t.name_ << endl;
	os << "  edges:  ";
	map<int, Edge>::const_iterator it;
	for (it = t.edges_.begin(); it != t.edges_.end(); ++it)
		os << it->second << "  ";
	os << endl;
	os << "  nodes:  ";
	for (set<int>::const_iterator n = t.nodes_.begin(); n != t.nodes_.end();
			++n)
		os << *n << "  ";
	os << endl;
	os << "  edges of node:" << endl;
	for (map<int, list<Edge> >::const_iterator n = t.edges_of_node_.begin();
			n != t.edges_of_node_.end(); ++n) {
		os << "    node: " << n->first << ": ";
		for (list<Edge>::const_iterator e = n->second.begin();
				e != n->second.end(); ++e)
			os << *e << "  ";
		os << endl;
	}
	return os;
}

} // namespace Reduze

